"""
Some explanations to let you understand it better:
Entrada (t): tiempo normalizado en [0, 1].
Salida (a): progreso no lineal en [0, 1].
"""
from yta_math_easings.abstract import EasingFunction
from yta_validation.parameter import ParameterValidator
from abc import ABC


class EasingAnimatedFunction(ABC):
    """
    *Abstract class*

    Abstract class to be inherited by the specific
    easing implementations.

    An easing function that is able to calculate a
    value according to the `n` parameter provided,
    but within a range of possible values and
    according to the duration set.

    The `t` value is representing the time lapsed
    for the easing function, as a normalized value
    in the `[0.0, 1.0]` range.

    The `a` value is representing the progress for
    the easing function, as a normalized value in
    the `[0.0, 1.0]` range.
    """
    _limit = (0, 1)
    """
    *For internal use only*

    The limit of the alpha value we can use.
    """

    def __init__(
        self,
        start_value: float,
        end_value: float,
        duration: float,
        easing_function: EasingFunction
    ):
        ParameterValidator.validate_mandatory_positive_number('duration', duration, do_include_zero = False)
        ParameterValidator.validate_mandatory_instance_of('easing_function', easing_function, EasingFunction)

        self.start_value: float = start_value
        """
        The start value of the valid range of the animated
        easing function.
        """
        self.end_value: float = end_value
        """
        The end value of the valid range of the animated
        easing function.
        """
        self.duration: float = duration
        """
        The total duration (in seconds) that the animated
        easing will take to go from the `start_value`
        to the `end_value`, that can be used to calculate
        the different values at the different time moments
        requested by the user.
        """
        self._easing_function: EasingFunction = easing_function
        """
        The easing function associated to this animated
        easing function, that will be the one used to
        calculate the values according to the definition.
        """
        
    def func(
        self,
        n: float,
        **kwargs
    ) -> float:
        """
        Alias for the `get_progress_from_t` function.
        """
        return self.get_progress_from_t(
            t = n,
            **kwargs
        )

    def get_progress_from_t(
        self,
        t: float,
        **kwargs
    ) -> float:
        """
        Also called `func`.

        Apply the easing to the `t` value provided, that is
        the time but normalized, so it must be a value in
        the `[0.0, 1.0]` range, and get the normalized
        `progress` associated to it based on the specific
        easing function you are using.

        The `t` represents the time but normalized, and the
        returned `a` will be the progress done in that period
        of time.
        """
        ParameterValidator.validate_mandatory_number_between(
            name = 't',
            value = t,
            lower_limit = 0.0,
            upper_limit = 1.0,
            do_include_lower_limit = True,
            do_include_upper_limit = True
        )

        return self._easing_function.func(
            t,
            **kwargs
        )
    
    def ease(
        self,
        alpha: float,
        **kwargs
    ) -> float:
        """
        Alias for the `get_t_from_progress` function.
        """
        return self.get_t_from_progress(
            progress = alpha,
            **kwargs
        )
    
    def get_t_from_progress(
        self,
        progress: float,
        **kwargs
    ) -> float:
        """
        Also called `ease` function.

        Get the `t` real time moment that is associated to
        the `progress` normalized value provided, according
        to the internal `start_value` and `end_value`.

        The `progress` must be a normalized value in the 
        `[0.0, 1.0]` range.
        """
        ParameterValidator.validate_mandatory_number_between(
            name = 'progress',
            value = progress,
            lower_limit = 0.0,
            upper_limit = 1.0,
            do_include_lower_limit = True,
            do_include_upper_limit = True
        )

        t = self._limit[0] * (1 - progress) + self._limit[1] * progress
        t /= self.duration
        progress = self.get_progress_from_t(
            t = t,
            **kwargs
        )

        return self.end_value * progress + self.start_value * (1 - progress)
    
    def __call__(self, alpha: float) -> float:
        """
        Special method to allow us doing
        `AnimatedEasingFunction.ease(alpha)` directly.
        """

        """
        permite escribir:
        value = animation(alpha)
        en lugar de:
        value = animation.ease(alpha)
        """
        return self.get_t_from_progress(alpha)





def get_class(
    name: str
    # TODO: Add the imports if the method is needed
) -> any:
# ) -> Union[EasingFunction, None]:
    """
    Get the easing function class that corresponds to the
    `name` provided, if existing.
    """
    from yta_math_easings.linear import LinearEasing

    # TODO: Add all the classes here
    classes = [
        LinearEasing
    ]

    for cls in classes:
        if cls.__name__.lower() == f'{name.lower()}easing':
            return cls
        
    return None


from yta_validation.parameter import ParameterValidator
from yta_programming.decorators.requires_dependency import requires_dependency
from abc import ABC, abstractmethod

import numpy as np


class EasingFunction(ABC):
    """
    *Abstract class*

    Abstract class to be inherited by the specific
    easing implementations.

    An easing function that is able to calculate a
    value according to the `n` parameter provided.
    """

    def func(
        self,
        t: float,
        **kwargs
    ) -> float:
        """
        Apply the easing to the `t` value provided, that must
        be a value in the `[0.0, 1.0]` range.
        """
        ParameterValidator.validate_mandatory_number_between(
            name = 't',
            value = t,
            lower_limit = 0.0,
            upper_limit = 1.0,
            do_include_lower_limit = True,
            do_include_upper_limit = True
        )

        return self._func(
            t = t,
            **kwargs
        )

    @abstractmethod
    def _func(
        self,
        t: float,
        **kwargs
    ) -> float:
        """
        *For internal use only*

        The proper calculation of the easing, that must be
        implemented for each specific class.
        """
        pass

    def __call__(
        self,
        t: float
    ) -> float:
        """
        Special method to allow us doing
        `EasingFunction.func(t)` directly.

        Apply the easing to the `t` value provided, that must
        be a value in the `[0.0, 1.0]` range.
        """
        return self.func(t)

    @requires_dependency('matplotlib', 'yta_math', 'matplotlib')
    def plot(
        self,
        number_of_samples: int = 100
    ):
        """
        Plot a graphic representing `number_of_samples`
        values, from `0.0` to `1.0`, to see how the easing
        function works.
        """
        import matplotlib.pyplot as plt
        
        # Limit and draw axis
        plt.xlim(0, 1)
        plt.ylim(0, 1)
        plt.axhline(0, color = 'black', linewidth = 1)
        plt.axvline(0, color = 'black', linewidth = 1)

        plt.grid(True)

        # We will draw 'n'' values between 0.0 and 1.1
        x_vals = np.linspace(0.0, 1.0, number_of_samples).tolist()
        y_vals = [
            self.get_n_value(x)
            for x in x_vals
        ]
        plt.scatter(x_vals, y_vals, color = 'white', edgecolors = 'black', s = 100)

        # TODO: I think we don't need to draw anything
        # else. Remove this below if it is ok.
        # # Draw points between nodes
        # for pair_of_node in self.pairs_of_nodes:
        #     positions = pair_of_node.get_n_xy_values_to_plot(100)
        #     t_xs, t_ys = zip(*positions)
        #     xs += t_xs
        #     ys += t_ys
       
        # plt.scatter(xs, ys, color = 'black', s = 1)
        
        plt.title('')
        plt.xlabel('X axis')
        plt.ylabel('Y axis')
        
        plt.show()



from yta_math_easings.abstract import EasingFunction

import numpy as np


class SlowIntoEasing(EasingFunction):
    """
    A slow into easing function.

    The formula:
    - `np.sqrt(1 - (1 - n) * (1 - n))`
    """

    def _func(
        self,
        n: float
    ) -> float:
        return np.sqrt(1 - (1 - n) * (1 - n))


from yta_math_easings.abstract import EasingFunction


class LinearEasing(EasingFunction):
    """
    A linear easing function.

    The formula:
    - `n`
    """

    def _func(
        self,
        n: float
    ) -> float:
        return n
    



"""
Module to include the functionality related
to easings.

Thanks for the inspiration:
- https://github.com/semitable/easing-functions/blob/master/easing_functions
"""
from yta_math_easings.linear import LinearEasing
from yta_math_easings.slow_into import SlowIntoEasing


__all__ = [
    'LinearEasing',
    'SlowIntoEasing'
]


from yta_math_easings.animated.abstract import EasingAnimatedFunction
from yta_math_easings.slow_into import SlowIntoEasing


class SlowIntoAnimatedEasing(EasingAnimatedFunction):
    """
    A slow into function but animated, including a
    range of values and a duration.
    """

    def __init__(
        self,
        start_value: float,
        end_value: float,
        duration: float
    ):
        super().__init__(
            start_value = start_value,
            end_value = end_value,
            duration = duration,
            easing_function = SlowIntoEasing()
        )


"""
Module to hold the easing functions but using
start and end values and a time duration.
"""
from yta_math_easings.animated.linear import LinearAnimatedEasing
from yta_math_easings.animated.slow_into import SlowIntoAnimatedEasing


__all__ = [
    'LinearAnimatedEasing',
    'SlowIntoAnimatedEasing'
]